import io
import logging
import re
import typing as t

from bs4 import BeautifulSoup

from mu import units
from mu.exceptions import MuError
from mu.formats.base.reader import BaseReader
from mu.utils import youtube

from .common import TYPE_ATTR

logger = logging.getLogger(__name__)


class HtmlReader(BaseReader):
    # pylint: disable=super-init-not-called
    def __init__(self, unit_html: BeautifulSoup) -> None:
        self.unit_html = unit_html

    def parse(self) -> t.Iterable[units.Unit]:
        """
        Parse the html content.

        The dispatch method is called recursively by the child `on_header` method.
        """
        yield from self.iter_units(self.unit_html)

    def iter_units(self, unit_html: BeautifulSoup) -> t.Iterable[units.Unit]:
        yield from super().dispatch(unit_html.name, unit_html)

    def on_header(self, unit_html: BeautifulSoup) -> t.Iterable[units.Unit]:
        """
        Parse `<h1>, ...<h6>` DOM elements.

        This method yields a single Collection for the current header. Headers from
        level n+1 will be added as children, provided they are direct children.

        This method is a little difficult to read. The problem with html headers is
        that they break the concept of parent -> child inclusion. So children in the
        sense of a course are actually siblings in the html world, and we need to figure
        out which ones are direct children of the current unit.
        """
        header_level = get_header_level(unit_html.name)
        assert header_level is not None

        # Create collection unit
        UnitClass = units.Course if header_level == 1 else units.Collection
        unit: units.Collection = UnitClass(
            attributes=get_data_attributes(unit_html),
            title=unit_html.string.strip(),
        )

        # Find children units.
        siblings_are_children = True
        for child_html in unit_html.find_next_siblings():
            if not getattr(child_html, "name"):
                # Ignore raw strings
                continue
            if child_header_level := get_header_level(child_html.name):
                # Header found: all other siblings are actually children of another unit
                if child_header_level < header_level:
                    # Child is actually a parent header: stop searching for children
                    # Parent header will be parsed in the parent call.
                    break
                elif child_header_level == header_level:
                    # Child is a header with the same level:
                    # Stop parsing and yield from a different parser.
                    break
                elif child_header_level == header_level + 1:
                    # Direct child -> will be appended to children
                    for child in self.iter_units(child_html):
                        unit.add_child(child)
                    # Other siblings are no longer children of this unit
                    siblings_are_children = False
                else:
                    # Child is a grand-child, so we ignore it
                    continue
            else:
                if siblings_are_children:
                    # Found a non-header unit: append to children
                    # (and concatenate RawHtml units in the process)
                    for child in self.iter_units(child_html):
                        if (
                            unit.children
                            and isinstance(unit.children[-1], units.RawHtml)
                            and isinstance(child, units.RawHtml)
                        ):
                            # Concatenate all RawHtml children
                            unit.children[-1].concatenate(child)
                        else:
                            # Append child
                            unit.add_child(child)

        # Yield current unit
        yield unit

    on_h1 = on_header
    on_h2 = on_header
    on_h3 = on_header
    on_h4 = on_header
    on_h5 = on_header
    on_h6 = on_header

    def on_section(self, unit_html: BeautifulSoup) -> t.Iterable[units.Unit]:
        """
        Parse `<section>` DOM elements.

        The parsing function that is called depends on the value of the
        "data-mu-type" attribute.
        """
        unit_type = unit_html.attrs.get(TYPE_ATTR)
        if unit_type == "mcq":
            # Multiple choice question
            yield from process_mcq(unit_html)
        elif unit_type == "video":
            yield from process_video(unit_html)
        elif unit_type == "ftq":
            # Free text question
            yield from process_ftq(unit_html)
        else:
            logger.warning("Unit type is unsupported by HTML reader: %s", unit_type)

    def _on_html(self, unit_html: BeautifulSoup) -> t.Iterable[units.Unit]:
        """
        All data-* attributes are copied to the RawHtml unit.
        """
        yield units.RawHtml(
            contents=str(unit_html),
            attributes=get_data_attributes(unit_html),
        )

    # Add here all html elements that should be converted to RawHtml
    on_div = _on_html
    on_p = _on_html
    on_pre = _on_html
    on_video = _on_html


class DocumentReader(HtmlReader):
    """
    Same as Reader, but build reader from top-level h1 header.
    """

    def __init__(self, document: BeautifulSoup) -> None:
        if h1 := document.find(name="h1"):
            super().__init__(h1)
        else:
            raise MuError("Could not find any h1 element in the HTML document")


class StringReader(DocumentReader):
    """
    Same as DocumentReader, but for html-formatted strings.

    Convenient for unit testing.
    """

    def __init__(self, contents: str) -> None:
        super().__init__(beautiful_soup(contents))


class Reader(StringReader):
    """
    Same as StringReader, but with a constructor that takes a file object as argument.
    """

    def __init__(self, path: str) -> None:
        with open(path, encoding="utf8") as f:
            super().__init__(f.read())


def get_header_level(h: str) -> t.Optional[int]:
    """
    Parse the header level: "hX" -> X

    Return None in case of no match.
    """
    match = re.match(r"h([1-9])", h)
    if not match:
        return None
    return int(match.group(1))


def get_data_attributes(unit_html: BeautifulSoup) -> t.Dict[str, str]:
    """
    Return all attributes that start with "data-"
    """
    return {
        key[5:]: value
        for key, value in unit_html.attrs.items()
        if key.startswith("data-")
    }


def process_mcq(unit_html: BeautifulSoup) -> t.Iterable[units.Unit]:
    """
    <ul> tags may contain multiple choice questions. In such cases, the first <li>
    is the question and each subsequent <li> element starts with either ✅ or ❌.
    """
    title, question, answers = get_question_answers(unit_html)
    right = "✅"
    wrong = "❌"
    evaluated_answers = []
    for answer in answers:
        is_correct = False
        if answer.startswith(right):
            is_correct = True
        elif answer.startswith(wrong):
            is_correct = False
        else:
            # Not a multiple choice question
            raise MuError(
                f"Incorrectly formatted answer in multiple choice question: "
                f"should start with either {right} or {wrong}"
            )
        evaluated_answers.append((answer[1:].strip(), is_correct))

    # Generate MCQ
    yield units.MultipleChoiceQuestion(
        title=title,
        question=question,
        answers=evaluated_answers,
    )


def process_ftq(unit_html: BeautifulSoup) -> t.Iterable[units.Unit]:
    title, question, answers = get_question_answers(unit_html)
    yield units.FreeTextQuestion(title=title, question=question, answers=answers)


def process_video(unit_html: BeautifulSoup) -> t.Iterable[units.Unit]:
    """
    https://developer.mozilla.org/en-US/docs/Web/HTML/Element/video
    """
    title = ""
    if title_html := find_title(unit_html):
        title = title_html.string

    video_html = unit_html.find("video")
    iframe_html = unit_html.find("iframe")
    if video_html is None and iframe_html is None:
        raise MuError(
            f"Missing <video> or <iframe> element in unit labelled as video: {unit_html}"
        )

    if video_html:
        sources: t.List[str] = []
        if src := video_html.attrs.get("src"):
            sources.append(src)
        for source_html in video_html.find_all("source"):
            if source := source_html.attrs.get("src"):
                if source not in sources:
                    sources.append(source)
        yield units.Video(title=title, sources=sources)
    else:
        src = iframe_html.attrs.get("src")
        if youtube_video_id := youtube.get_embed_video_id(src):
            yield units.Video(
                title=title, sources=[youtube.get_watch_url(youtube_video_id)]
            )


def get_question_answers(unit_html: BeautifulSoup) -> t.Tuple[str, str, t.List[str]]:
    title = ""
    if title_html := find_title(unit_html):
        title = title_html.string

    question: t.Optional[str] = None
    answers = []
    question_html = unit_html.find("p")
    if question_html is None:
        raise MuError(f"Missing <p> element in multiple choice question: {unit_html}")
    question = question_html.string.strip()
    if question is None:
        raise MuError(
            f"Missing question string in multiple choice question: {unit_html}"
        )
    ul_html = unit_html.find("ul")
    if ul_html is None:
        raise MuError(f"Missing <ul> element in multiple choice question: {unit_html}")
    for li_html in ul_html.find_all("li"):
        answer = li_html.string.strip()
        answers.append(answer.strip())

    return title, question, answers


def find_title(unit_html: BeautifulSoup) -> t.Optional[BeautifulSoup]:
    return unit_html.find(
        [
            "h1",
            "h2",
            "h3",
            "h4",
            "h5",
            "h6",
        ]
    )


def beautiful_soup(src: t.Union[str, io.TextIOBase]) -> BeautifulSoup:
    return BeautifulSoup(src, "html5lib")



import logging
import os

from bs4 import BeautifulSoup
from bs4.element import Tag

from mu import units
from mu.formats.base.writer import BaseWriter
from mu.utils import youtube

from .common import TYPE_ATTR

logger = logging.getLogger(__name__)


class UnstyledWriter(BaseWriter):
    def __init__(self) -> None:
        self.document = beautiful_soup("<!DOCTYPE html>")

    def write_to(self, path: str) -> None:
        with open(path, "w", encoding="utf-8") as f:
            f.write(self.dumps())

    def dumps(self) -> str:
        # Convenience method
        contents: str = self.document.decode(pretty_print=True)
        return contents

    def append_to_body(self, tag: Tag) -> None:
        self.document.html.body.append(tag)

    def get_header(self, unit: units.Unit) -> Tag:
        tag = Tag(
            name=f"h{unit.depth + 1}",
            attrs={f"data-{k}": v for k, v in unit.attributes.items()},
        )
        if title := unit.title:
            tag.string = title
        return tag

    def on_collection(self, unit: units.Collection) -> None:
        self.append_to_body(self.get_header(unit))

    def on_course(self, unit: units.Course) -> None:
        self.append_to_body(self.get_header(unit))

    def on_multiplechoicequestion(self, unit: units.MultipleChoiceQuestion) -> None:
        section_html = Tag(name="section", attrs={TYPE_ATTR: "mcq"})

        # Write title
        section_html.append(self.get_header(unit))

        # Write question
        question_html = Tag(name="p")
        question_html.string = unit.question
        section_html.append(question_html)
        answers_html = Tag(name="ul")

        # Write answers
        for answer, is_correct in unit.answers:
            answer_html = Tag(name="li")
            answer_html.string = f"{'✅' if is_correct else '❌'} {answer}"
            answers_html.append(answer_html)
        section_html.append(answers_html)

        self.append_to_body(section_html)

    def on_freetextquestion(self, unit: units.FreeTextQuestion) -> None:
        section_html = Tag(name="section", attrs={TYPE_ATTR: "ftq"})

        # Write title
        section_html.append(self.get_header(unit))

        # Write question
        question_html = Tag(name="p")
        question_html.string = unit.question
        section_html.append(question_html)
        answers_html = Tag(name="ul")

        # Write answers
        for answer, _ in unit.answers:
            answer_html = Tag(name="li")
            answer_html.string = answer
            answers_html.append(answer_html)
        section_html.append(answers_html)

        self.append_to_body(section_html)

    def on_video(self, unit: units.Video) -> None:
        """
        We parse the video sources. If one is youtube, we include a youtube iframe.
        Else, we include a <video> element.
        """
        section_html = Tag(name="section", attrs={TYPE_ATTR: "video"})

        # Write title
        section_html.append(self.get_header(unit))

        # TODO handle transcripts
        video_tag = get_video_tag(unit.sources)
        section_html.append(video_tag)
        self.append_to_body(section_html)

    def on_rawhtml(self, unit: units.RawHtml) -> None:
        self.append_to_body(BeautifulSoup(unit.contents, "html.parser"))


class Writer(UnstyledWriter):
    """
    Add some basic CSS styling to the default writer.
    """

    def on_course(self, unit: units.Course) -> None:
        super().on_course(unit)
        css = Tag(name="style")
        css.string = """
body {
    max-width: 1024px;
    margin: auto;
    font-family: sans-serif;
}

video {
    width: 800px;
}

iframe {
    width: 800px;
    height: 450px;
}
"""
        self.document.html.head.append(css)


def get_video_tag(sources: list[str]) -> BeautifulSoup:
    """
    Return either a <video> or <iframe> element based on the video sources.
    """
    video_tag = Tag(name="video", attrs={"controls": None})
    for source in sources:
        if youtube_embed_url := youtube.get_embed_url(source):
            # Video url points to youtube. In such a case, we don't return a <video> but an iframe.
            return Tag(
                name="iframe",
                attrs={"src": youtube_embed_url},
            )
        # Video is standard html5
        video_extension = os.path.splitext(source)[1].lower()
        video_type = {
            ".mp4": "mp4",
            ".mov": "mp4",
            ".ogg": "ogg",
            ".webm": "webm",
        }.get(video_extension)
        if video_type:
            video_tag.append(
                Tag(
                    name="source",
                    attrs={"src": source, "type": f"video/{video_type}"},
                )
            )
        else:
            logger.warning("Unsupported video extension: %s", video_extension)
    return video_tag


def beautiful_soup(contents: str) -> BeautifulSoup:
    return BeautifulSoup(contents, "html5lib")



#! /usr/bin/env python
import argparse
import importlib
import logging
import os
import sys
import typing as t

from mu.exceptions import MuError
from mu.formats.base.reader import BaseReader
from mu.formats.base.writer import BaseWriter

logger = logging.getLogger(__name__)


def main() -> None:
    try:
        run()
    except MuError as e:
        logger.error(e.args[0])
        sys.exit(1)


def run() -> None:
    args = parse_args()

    ReaderClass: t.Type[BaseReader] = import_class(
        f"mu.formats.{args.from_format}.reader", "Reader"
    )
    WriterClass: t.Type[BaseWriter] = import_class(
        f"mu.formats.{args.to_format}.writer", "Writer"
    )

    try:
        course = ReaderClass(args.input).read()
    except MuError as e:
        raise MuError(f"Error in {args.from_format} reader: {e.args[0]}") from e
    try:
        WriterClass().write(course).write_to(args.output)
    except MuError as e:
        raise MuError(f"Error in in {args.to_format} writer: {e.args[0]}") from e


def import_class(module_name: str, class_name: str) -> t.Any:
    module = importlib.import_module(module_name)
    return getattr(module, class_name)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Convert online courses from one format to another."
    )
    parser.add_argument(
        "-f", "--from-format", choices=["html", "md", "olx"], default=None
    )
    parser.add_argument(
        "-t", "--to-format", choices=["html", "md", "olx"], default=None
    )
    parser.add_argument("-v", "--verbose", action="count", default=0)
    parser.add_argument("input", help="Input file or directory.")
    parser.add_argument("output", help="Output file or directory.")
    args = parser.parse_args()

    # Configure logging
    log_level: int = logging.WARNING
    if args.verbose == 1:
        log_level = logging.INFO
    elif args.verbose >= 2:
        log_level = logging.DEBUG
    logging.basicConfig(
        level=log_level,
        format="[%(asctime)s][%(name)s] %(levelname)s: %(message)s",
        force=True,
    )

    # Attempt to guess input/output file format
    if args.from_format is None:
        if args.input.lower().endswith(".html"):
            args.from_format = "html"
        elif args.input.lower().endswith(".md"):
            args.from_format = "md"
        elif os.path.isdir(args.input):
            args.from_format = "olx"
        else:
            raise MuError("Could not detect input file format.")
        logger.info("Detected input format: %s", args.from_format)

    if args.to_format is None:
        if args.output.lower().endswith(".html"):
            args.to_format = "html"
        elif args.output.lower().endswith(".md"):
            args.to_format = "md"
        elif os.path.isdir(args.output):
            args.to_format = "olx"
        else:
            raise MuError("Could not detect output file format.")
        logger.info("Detected output format: %s", args.to_format)

    return args


if __name__ == "__main__":
    main()



import hashlib
import logging
import os
import typing as t

from bs4 import BeautifulSoup, Tag

from mu import units
from mu.exceptions import MuError
from mu.formats.base.writer import BaseWriter
from mu.utils.youtube import get_video_id as get_youtube_video_id

logger = logging.getLogger(__name__)


class Writer(BaseWriter):
    def __init__(self) -> None:
        # List of (xml, parent_xml, path) tuples
        self.xml_paths: t.List[t.Tuple[Tag, str]] = []
        # unit -> xml dict
        self.unit_xml: t.Dict[units.Unit, Tag] = {}

    def write_to(self, path: str) -> None:
        for unit_xml, unit_path in self.xml_paths:
            # Write all xml files
            write_xml(unit_xml, os.path.join(path, unit_path), makedirs=True)

    def on_collection(self, unit: units.Collection) -> None:
        self.process_top_level_unit(unit)

    def on_course(self, unit: units.Course) -> None:
        # course.xml
        attributes = {
            "org": "organization",
            "course": "course",
            "url_name": get_url_name(unit),
        }
        for name in ["org", "course", "url_name"]:
            # TODO IMPORTANT should we also copy all the olx-* attributes? (I think we should)
            # TODO IMPORTANT should we also copy the attributes when we convert from olx?
            if olx_attr := unit.attributes.get(f"olx-{name}"):
                attributes[name] = olx_attr
            else:
                logger.warning(
                    "Top-level course does not have required attribute 'olx-%s'. Will default to '%s'.",
                    name,
                    attributes[name],
                )
        unit_xml = Tag(name="course", attrs=attributes)
        self.xml_paths.append((unit_xml, "course.xml"))
        self.unit_xml[unit] = unit_xml

        # course/<title>.xml
        self.process_top_level_unit(unit)

    def on_multiplechoicequestion(self, unit: units.MultipleChoiceQuestion) -> None:
        """
        https://edx.readthedocs.io/projects/edx-open-learning-xml/en/latest/problem-xml/checkbox.html

        Note that we generate checkbox problems by default, and not multiple choice problem, which support only a single answer:
        https://edx.readthedocs.io/projects/edx-open-learning-xml/en/latest/problem-xml/multiple_choice.html#multiple-choice-problem-olx-reference
        """
        problem_xml = self.process_unit(unit, "problem")
        response_xml = Tag(name="choiceresponse")
        question_xml = Tag(name="label")
        question_xml.string = unit.question
        response_xml.append(question_xml)
        responsegroup_xml = Tag(name="checkboxgroup")
        for index, (answer, is_correct) in enumerate(unit.answers):
            answer_xml = Tag(
                name="choice",
                attrs={"correct": str(is_correct).lower(), "name": str(index)},
            )
            answer_xml.string = answer
            responsegroup_xml.append(answer_xml)

        response_xml.append(responsegroup_xml)
        problem_xml.append(response_xml)

    def on_freetextquestion(self, unit: units.MultipleChoiceQuestion) -> None:
        """
        https://edx.readthedocs.io/projects/edx-open-learning-xml/en/latest/problem-xml/text_input.html
        """
        problem_xml = self.process_unit(unit, "problem")
        response_xml = Tag(name="stringresponse")
        response_xml.attrs["answer"] = unit.answers[0][0]
        question_xml = Tag(name="label")
        question_xml.string = unit.question
        response_xml.append(question_xml)
        for answer, _ in unit.answers[1:]:
            answer_xml = Tag(
                name="additional_answer",
                attrs={"answer": answer},
            )
            response_xml.append(answer_xml)
        problem_xml.append(response_xml)

    def on_rawhtml(self, unit: units.RawHtml) -> None:
        """
        https://edx.readthedocs.io/projects/edx-open-learning-xml/en/latest/components/html-components.html

        Note that the spec is incorrect. We should use "filename" and not "file_name".
        """
        html_xml = self.process_unit(unit, "html")

        # Create a <url_name>.html file with the html contents
        # Note the parser we use to prevent adding <html> tags
        # https://beautiful-soup-4.readthedocs.io/en/latest/#specifying-the-parser-to-use
        url_name = get_url_name(unit)
        self.xml_paths.append(
            (
                BeautifulSoup(unit.contents, "html.parser"),
                os.path.join("html", f"{url_name}.html"),
            )
        )

        # Point to the html file
        html_xml.attrs["filename"] = url_name

    def on_video(self, unit: units.Video) -> None:
        """
        https://edx.readthedocs.io/projects/edx-open-learning-xml/en/latest/components/video-components.html
        """
        video_xml = self.process_unit(unit, "video")
        # We need to define an empty youtube url; otherwise the default video with Anant is picked up.
        video_xml.attrs["youtube_id_1_0"] = ""
        for source in unit.sources:
            if youtube_video_id := get_youtube_video_id(source):
                video_xml.attrs["youtube_id_1_0"] = youtube_video_id
            else:
                video_xml.append(Tag(name="source", attrs={"src": source}))

    def process_top_level_unit(self, unit: units.Unit) -> None:
        """
        Process title units.
        """
        olx_types = ["course", "chapter", "sequential", "vertical"]
        unit_depth = unit.depth
        if unit_depth >= len(olx_types):
            # The title unit will not be created. Instead, we will use its title as the
            # unit display name.
            return
        if unit_depth >= len(olx_types) + 1:
            logger.warning(
                "Cannot render top-level unit '%s' of depth %d in olx.",
                unit.title,
                unit_depth,
            )
            return
        unit_type = olx_types[unit_depth]
        self.process_unit(unit, unit_type)

    def process_unit(self, unit: units.Unit, unit_type: str) -> Tag:
        """
        - If the unit has no url_name, generate one
        - Create the xml object that corresponds to <unit type>/<url_name>.xml
        - Append some xml reference to the parent xml
        - If children don't have a url_name, generate it
        """
        # Get/Generate url_name
        url_name = get_url_name(unit)

        # <unit_type>/<url_name>.xml
        display_name = unit.title
        if not display_name and unit.parent:
            # Borrow the title from the above unit
            display_name = unit.parent.title
        unit_xml = Tag(name=unit_type, attrs={"display_name": display_name})
        self.xml_paths.append(
            (
                unit_xml,
                os.path.join(unit_type, f"{url_name}.xml"),
            )
        )
        self.unit_xml[unit] = unit_xml

        # Find the nearest parent for which we have created an xml element
        parent = unit.parent
        while parent is not None:
            # Append <type url_name="..."> to that parent (if any)
            if parent_xml := self.unit_xml.get(parent):
                parent_xml.append(Tag(name=unit_type, attrs={"url_name": url_name}))
                break
            parent = parent.parent

        return unit_xml


def get_url_name(unit: units.Unit) -> str:
    if url_name := unit.attributes.get("olx-url_name"):
        return url_name
    url_name_hash = get_url_name_hash(unit)
    return hashlib.md5(url_name_hash.encode()).hexdigest()


def get_url_name_hash(unit: units.Unit) -> str:
    # Generate a repeatable hash based on the unit position in the tree
    # E.g: 3_11_0_5
    current_unit = unit
    unit_hash = ""
    while current_unit.parent is not None:
        if unit_hash:
            unit_hash = f"_{unit_hash}"
        nth_child = current_unit.parent.children.index(current_unit)
        unit_hash = f"{nth_child}{unit_hash}"
        current_unit = current_unit.parent
    return unit_hash


def write_xml(tag: Tag, path: str, makedirs: bool = False) -> None:
    directory = os.path.dirname(path)
    if makedirs:
        os.makedirs(directory, exist_ok=True)
    if not os.path.exists(directory):
        raise MuError(f"Destination directory does not exist: {directory}")
    with open(path, "w", encoding="utf-8") as f:
        f.write(tag.prettify())
        f.write("\n")



import re
import typing as t


def is_video_url(url: str) -> bool:
    return get_video_id(url) is not None


def is_embed_url(url: str) -> bool:
    return get_embed_video_id(url) is not None


def get_watch_url(youtube_id: str) -> str:
    return f"https://www.youtube.com/watch?v={youtube_id}"


def get_embed_url(url: str) -> t.Optional[str]:
    video_id = get_video_id(url)
    if video_id is None:
        return None
    return f"https://www.youtube.com/embed/{video_id}"


def get_video_id(url: str) -> t.Optional[str]:
    match = re.match(
        r"https?://(www\.)?(youtube\.com/watch\?v=|youtu.be/)(?P<id>[a-zA-Z0-9_]+)(&.*)?$",
        url,
    )
    return None if match is None else match.group("id")


def get_embed_video_id(url: str) -> t.Optional[str]:
    match = re.match(
        r"https?://(www\.)?youtube\.com/embed/(?P<id>[a-zA-Z0-9_]+)(&.*)?$",
        url,
    )
    return None if match is None else match.group("id")



import subprocess
import tempfile

from mu.formats.html.writer import Writer as HtmlWriter


class Writer(HtmlWriter):
    """
    Same as HTML writer, except that we convert the output to Markdown.
    """

    def write_to(self, path: str) -> None:
        # Write html to temporary file
        with tempfile.NamedTemporaryFile("w") as of:
            super().write_to(of.name)
            of.flush()
            # Convert file with pandoc
            subprocess.check_call(
                ["pandoc", "--from=html", "--to=markdown", of.name, f"--output={path}"]
            )



import shutil
import subprocess
import tempfile

from mu.exceptions import MuError
from mu.formats.html.reader import DocumentReader, beautiful_soup


class Reader(DocumentReader):
    def __init__(self, path: str) -> None:
        """
        Same as an HTML reader, but convert from markdown at runtime.
        """
        # Convert markdown to html with Pandoc
        try:
            html_content = subprocess.check_output(
                ["pandoc", "--from=markdown", "--to=html5", path]
            ).decode()
        except FileNotFoundError as e:
            if shutil.which("pandoc") is None:
                raise MuError("pandoc must be installed to read from Markdown.") from e

        # Load html
        html = beautiful_soup(html_content)

        # Trim all <p> tags surrounding videos
        # Note: as far as I know, we no longer need to do this
        # for video_html in html.find_all("video"):
        #     if video_html.parent and video_html.parent.name == "p":
        #         video_html.parent.replace_with_children()

        super().__init__(html)


class StringReader(Reader):
    """
    Same as reader, but read from raw string. Convenient for unit testing.
    """

    def __init__(self, content: str) -> None:
        with tempfile.NamedTemporaryFile("w") as of:
            of.write(content)
            of.flush()
            super().__init__(of.name)



import typing as t

U = t.TypeVar("U", bound="Unit")


class Unit:
    """
    A generic course unit.

    All units have an optional title and key/value attributes.

    Courses follow a tree structure. Units which are not containers are terminal leaves.
    Every unit (except the top-level one) has a parent which is an instance of a
    Collection.
    """

    def __init__(
        self, attributes: t.Optional[t.Dict[str, str]] = None, title: str = ""
    ):
        self.attributes = attributes or {}
        self.parent: t.Optional["Collection"] = None
        self.title = title

    @property
    def depth(self) -> int:
        if self.parent is None:
            return 0
        return self.parent.depth + 1


class Collection(Unit):
    """
    A special type of Unit which can include children units.
    """

    def __init__(
        self, attributes: t.Optional[t.Dict[str, str]] = None, title: str = ""
    ):
        super().__init__(attributes=attributes, title=title)
        self.children: t.List[Unit] = []

    def add_child(self, unit: U) -> U:
        unit.parent = self
        self.children.append(unit)
        return unit


class Course(Collection):
    """
    Top-level element of a course.

    For now there is nothing special about this unit, but we may add extra properties in
    the future.
    """


class MultipleChoiceQuestion(Unit):
    def __init__(
        self,
        attributes: t.Optional[t.Dict[str, str]] = None,
        title: str = "",
        question: str = "",
        answers: t.Optional[t.List[t.Tuple[str, bool]]] = None,
    ):
        super().__init__(attributes=attributes, title=title)
        self.question = question
        answers = answers or []
        self.answers = [(answer.strip(), is_correct) for answer, is_correct in answers]


class FreeTextQuestion(MultipleChoiceQuestion):
    """
    A question for which the student is presented with a free text field to answer.

    Multiple answers are usually accepted.
    """

    def __init__(
        self,
        attributes: t.Optional[t.Dict[str, str]] = None,
        title: str = "",
        question: str = "",
        answers: t.Optional[t.List[str]] = None,
    ):
        answers = answers or []
        super().__init__(
            attributes=attributes,
            title=title,
            question=question,
            answers=[(answer, True) for answer in answers],
        )


class RawHtml(Unit):
    def __init__(
        self,
        attributes: t.Optional[t.Dict[str, str]] = None,
        title: str = "",
        contents: str = "",
    ):
        super().__init__(attributes=attributes, title=title)
        self.contents = contents.strip()

    def concatenate(self, unit: "RawHtml") -> "RawHtml":
        self.contents += "\n" + unit.contents
        return self


class Video(Unit):
    def __init__(
        self,
        attributes: t.Optional[t.Dict[str, str]] = None,
        title: str = "",
        sources: t.Optional[t.List[str]] = None,
    ):
        """
        Sources are urls to the video. They can be completely arbitrary, such as youtube
        links or mp4 files.
        """
        super().__init__(attributes=attributes, title=title)
        self.sources = sources or []



import typing as t

from mu import units
from mu.exceptions import MuError

T = t.TypeVar("T")


class BaseReader:
    """
    Abstract base class for reading course documents.

    Child classes should implement:
    - a constructor that takes a single path as argument.
    - the `parse()` method, which should yield the course units.
    """

    def __init__(self, path: str) -> None:
        raise NotImplementedError

    def read(self) -> units.Course:
        """
        Read the course content from the path.

        For now, only a single course is returned. In theory nothing would prevent us
        from iterating over multiple courses here. Maybe we'll do that in the future.
        """
        for course in self.parse():
            if not isinstance(course, units.Course):
                raise MuError(
                    f"Reader failed to read course. Expected {units.Course} object, but reader yielded {course.__class__}."
                )
            return course
        raise MuError("Reader failed to read course: no course found")

    def parse(self) -> t.Iterable[units.Unit]:
        """
        This is where the bulk of the parsing should happen.

        Parse the content of the file(s) and yield Unit objects.
        """
        raise NotImplementedError

    def dispatch(
        self, name: str, *args: t.Any, **kwargs: t.Any
    ) -> t.Iterable[units.Unit]:
        """
        Convenient utility method, which will call the `on_<name>` method from the
        reader with `*args, **kwargs`.

        If no corresponding method is found, don't do anything.
        """
        # Parse element itself
        on_func: t.Optional[t.Callable[[t.Any], t.Iterable[units.Unit]]] = getattr(
            self, f"on_{name}", None
        )
        if on_func is not None:
            yield from on_func(*args, **kwargs)



import io
import os
import typing as t

from setuptools import find_packages, setup

HERE = os.path.abspath(os.path.dirname(__file__))


def load_readme() -> str:
    with io.open(os.path.join(HERE, "README.md"), "rt", encoding="utf8") as f:
        return f.read()


def load_requirements(filename: str) -> t.List[str]:
    with io.open(
        os.path.join(HERE, "requirements", filename), "rt", encoding="utf-8"
    ) as f:
        return [line.strip() for line in f if is_requirement(line)]


def is_requirement(line: str) -> bool:
    return not (line.strip() == "" or line.startswith("#"))


setup(
    name="mu-courses",
    version="0.1.1",
    url="https://github.com/overhangio/mu/",
    project_urls={
        "Code": "https://github.com/overhangio/mu",
        "Issue tracker": "https://github.com/overhangio/mu/issues",
    },
    license="AGPLv3",
    author="Overhang.IO",
    author_email="contact@overhang.io",
    description="Course authoring for humans",
    long_description=load_readme(),
    long_description_content_type="text/markdown",
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
    python_requires=">=3.7",
    install_requires=load_requirements("base.in"),
    entry_points={"console_scripts": ["mu=mu.main:main"]},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    test_suite="tests",
)



import io
import logging
import os
import re
import typing as t

from bs4 import BeautifulSoup

from mu import units
from mu.exceptions import MuError
from mu.formats.base.reader import BaseReader
from mu.utils import youtube

logger = logging.getLogger(__name__)


class InlineReader(BaseReader):
    # pylint: disable=super-init-not-called
    def __init__(self, unit_xml: BeautifulSoup) -> None:
        self.unit_xml = unit_xml

    def get_child_reader(self, child_xml: BeautifulSoup) -> "InlineReader":
        return InlineReader(child_xml)

    def parse(self) -> t.Iterable[units.Unit]:
        # We skip name-less children, such as 'NavigableString'
        if not getattr(self.unit_xml, "name", None):
            return

        # Dispatch call to on_* functions
        yield from self.dispatch(self.unit_xml.name, self.unit_xml)

    def parse_children(self) -> t.Iterable[units.Unit]:
        # Parse children
        for child_xml in self.unit_xml.children:
            reader = self.get_child_reader(child_xml)
            yield from reader.parse()

    def _on_collection(
        self, unit_xml: BeautifulSoup, collection: t.Optional[units.Collection] = None
    ) -> t.Iterable[units.Unit]:
        """
        Dispatch function for course, chapter, sequential and vertical units.
        """
        if collection is None:
            collection = units.Collection(
                get_unit_attributes(unit_xml),
                title=unit_xml.attrs.get("display_name", ""),
            )
        for child in self.parse_children():
            collection.add_child(child)
        yield collection

    def on_course(self, unit_xml: BeautifulSoup) -> t.Iterable[units.Unit]:
        course = units.Course(
            get_unit_attributes(unit_xml), title=unit_xml.attrs.get("display_name", "")
        )
        yield from self._on_collection(unit_xml, course)

    on_chapter = _on_collection
    on_sequential = _on_collection
    on_vertical = _on_collection

    def on_problem(self, unit_xml: BeautifulSoup) -> t.Iterable[units.Unit]:
        """
        https://edx.readthedocs.io/projects/edx-open-learning-xml/en/latest/problem-xml/checkbox.html
        https://edx.readthedocs.io/projects/edx-open-learning-xml/en/latest/problem-xml/text_input.html
        """
        # Parse question
        question_xml = unit_xml.find("label")
        question = question_xml.string if question_xml else ""

        if response_xml := unit_xml.find("choiceresponse"):
            # Multiple choice question
            yield units.MultipleChoiceQuestion(
                title=unit_xml.attrs.get("display_name", ""),
                question=question,
                answers=[
                    (
                        answer_xml.string,
                        answer_xml.attrs.get("correct", "").lower() == "true",
                    )
                    for answer_xml in response_xml.find_all("choice")
                ],
            )
        elif response_xml := unit_xml.find("stringresponse"):
            # Free text question
            ftq_answers = []
            ftq_answers.append(response_xml.attrs["answer"])
            for answer_xml in response_xml.find_all("additional_answer"):
                ftq_answers.append(answer_xml.attrs["answer"])
            yield units.FreeTextQuestion(
                title=unit_xml.attrs.get("display_name", ""),
                question=question,
                answers=ftq_answers,
            )

    def on_html(self, unit_xml: BeautifulSoup) -> t.Iterable[units.Unit]:
        """
        https://edx.readthedocs.io/projects/edx-open-learning-xml/en/latest/components/html-components.html
        """
        contents = "\n".join([str(c) for c in unit_xml.contents])
        yield units.RawHtml(
            title=unit_xml.attrs.get("display_name", ""), contents=contents
        )

    def on_video(self, unit_xml: BeautifulSoup) -> t.Iterable[units.Unit]:
        """
        https://edx.readthedocs.io/projects/edx-open-learning-xml/en/latest/components/video-components.html
        """
        sources: t.List[str] = []
        if youtube_id := unit_xml.attrs.get("youtube_id_1_0"):
            sources.append(youtube.get_watch_url(youtube_id))
        elif youtube_id := unit_xml.attrs.get("youtube"):
            # Format: 1.00:<id>,2.5:<id>,...
            # We only look for the single speed video
            # Some videos only include a 'youtube' attribute, but no 'youtube_id_1_0'.
            if youtube_id_match := re.match(r"1(\.00)?:(?P<id>.+)", youtube_id):
                sources.append(youtube.get_watch_url(youtube_id_match.group("id")))
        for source_xml in unit_xml.find_all("source"):
            if source := source_xml.attrs.get("src"):
                sources.append(source)
        yield units.Video(
            get_unit_attributes(unit_xml),
            title=unit_xml.attrs.get("display_name", ""),
            sources=sources,
        )


class StringReader(InlineReader):
    """
    Reader for xml-formatted strings.

    Convenient for unit testing.
    """

    def __init__(self, contents: str) -> None:
        super().__init__(beautiful_soup(contents).find())


class Reader(InlineReader):
    """
    This reader can load xml files indicated by the `url_name` property. In such
    cases, the "url_name" will be stored among the unit attributes.
    """

    def __init__(
        self, root_directory: str, unit_xml: t.Optional[BeautifulSoup] = None
    ) -> None:
        self.root_directory = root_directory
        if unit_xml is None:
            # Load course.xml from file
            unit_xml = load_xml(os.path.join(self.root_directory, "course.xml"))
            if not hasattr(unit_xml, "course"):
                raise MuError(
                    "Badly formatted course.xml file: missing <course> element"
                )
            unit_xml = unit_xml.course

        # Load xml from file pointed by url_name
        if url_name := getattr(unit_xml, "attrs", {}).get("url_name"):
            url_name_path = os.path.join(
                self.root_directory, unit_xml.name, f"{url_name}.xml"
            )
            if not os.path.exists(url_name_path):
                logger.warning(
                    "Failed to load unit. File does not exist: %s", url_name_path
                )
            else:
                new_unit_xml = getattr(load_xml(url_name_path), unit_xml.name)
                if new_unit_xml is None:
                    raise MuError(
                        f"Element with name '{unit_xml.name}' could not be found in {url_name_path}"
                    )
                new_unit_xml.attrs["url_name"] = url_name
                # Copy attributes from source element
                new_unit_xml.attrs.update(unit_xml.attrs)
                # Replace source element
                unit_xml = new_unit_xml

        super().__init__(unit_xml)

    def get_child_reader(self, child_xml: BeautifulSoup) -> "Reader":
        return Reader(self.root_directory, child_xml)


def get_unit_attributes(unit_xml: BeautifulSoup) -> t.Dict[str, t.Any]:
    """
    Call this function in your custom `on_*` methods to get a dict of OLX key/value
    attributes to be added to the created unit.
    """
    attributes = {}
    for k, v in unit_xml.attrs.items():
        attributes[f"olx-{k}"] = v
    attributes["olx-type"] = unit_xml.name
    return attributes


def load_xml(path: str) -> BeautifulSoup:
    if not os.path.isfile(path):
        raise MuError(f"Missing XML file: '{path}'")
    with open(path, encoding="utf-8") as f:
        return beautiful_soup(f)


def beautiful_soup(src: t.Union[str, io.TextIOBase]) -> BeautifulSoup:
    return BeautifulSoup(src, "xml")



from mu import units


class BaseWriter:
    def write_to(self, path: str) -> None:
        raise NotImplementedError

    def write(self, unit: units.Unit) -> "BaseWriter":
        # Get the "on_<Name>" function that corresponds to the unit type.
        on_func = getattr(self, f"on_{unit.__class__.__name__.lower()}")
        on_func(unit)

        # Write children recursively: depth-first traversal
        if isinstance(unit, units.Collection):
            for child in unit.children:
                self.write(child)

        return self

    def on_collection(self, unit: units.Collection) -> None:
        pass

    def on_course(self, unit: units.Course) -> None:
        pass

    def on_multiplechoicequestion(self, unit: units.MultipleChoiceQuestion) -> None:
        pass

    def on_video(self, unit: units.Video) -> None:
        pass

    def on_rawhtml(self, unit: units.RawHtml) -> None:
        pass
